// This class provides a generic bit field implementation for an 
// arbitrary number of bits. Based on code written by irfan.hamid@gmail.com
//
// Implementation notes:
// // The bit field is implemented as an array of unsigned int that is
// allocated at object creation.
//
// field: The array that represents the bit field itself
// bit_count: The total number of bits in the bit field

#ifndef __GBITFIELD_H__
#define __GBITFIELD_H__
#include "GBase.h"
#define SZ_UINT sizeof(unsigned int)

class GBitField {
  protected:
    unsigned int *field;
    unsigned int bit_count;

  public:
    // The constructor takes an int param which gives the number of
    // bits in this bit field.
    GBitField (int bc) {
      bit_count = bc;
      GMALLOC(field, ceil(bc/SZ_UINT*8));
      }
    ~GBitField () {
      GFREE(field);
      }

    // This function sets the corresponding bit in the field equal to v
    // (or 1, if v was not given)
    // Returns: 0 on success, -1 on error.
    int set(int bit, int v=1) {
      // Sanity check
      if (bit >= bit_count || !field)
        return -1;
      // The correct index into the vector will be given by bit/SZ_UNIT. The
      // index into the correct vector element is given by bit%SZ_UNIT, to
      // achieve this, simply left shift 0x01 the appropriate number of times.
      if (v==0) {
        field[bit/SZ_UINT] &= ~(1 << (bit%(SZ_UINT*8)-1));
        }
      else {  
        field[bit/SZ_UINT] |= (1 << (bit%(SZ_UINT*8)-1));
        }
      return 0;
      }
      
    // This function returns the value of the corresponding bit.
    // Returns: The value of the bit, if the field is initialized and bit index is
    // within bounds, -1 otherwise.
    int get(int bit) {
      if (bit >= bit_count || !field)
        return -1;
      return (field[bit/SZ_UINT] & (0x00000001 << (bit%(SZ_UINT*8)-1)) ? 1 : 0);
      }

    // This function sets the corresponding bit in the field equal to 0.
    // Returns: 0 on success, -1 on error.
    int reset(int bit) {
      if (bit >= bit_count || !field)
        return -1;
      field[bit/SZ_UINT] &= ~(0x00000001 << (bit%(SZ_UINT*8)-1));
      return 0;
      }
};



#endif
