#include <cstdlib>
#include <cstdio>
#include <string>

#include "bam/bam.h"
#include "bam/sam.h"

using namespace std;

extern unsigned short bam_char2flag_table[];

uint8_t* realloc_bdata(bam1_t *b, int size) {
  if (b->m_data < size) {
        b->m_data = size;
        kroundup32(b->m_data);
        b->data = (uint8_t*)realloc(b->data, b->m_data);
        b->data_len=size;
        }
  return b->data;
}

uint8_t* dupalloc_bdata(bam1_t *b, int size) {
  //same as realloc_bdata, but does not free previous data
  //but returns it instead
  //it ALWAYS duplicates data
  b->m_data = size;
  kroundup32(b->m_data);
  uint8_t* odata=b->data;
  b->data = (uint8_t*)malloc(b->m_data);
  memcpy((void*)b->data, (void*)odata, b->data_len);
  b->data_len=size;
  return odata; //user must FREE this after
}


class GBamRecord {
   bam1_t* b;
   // b->data has the following strings concatenated:
   //  qname (including the terminal \0)
   //  +cigar (each event encoded on 32 bits)
   //   +seq  (4bit-encoded)
   //    +qual
   //     +aux
   bool novel;
 public:
   GBamRecord(bam1_t* from_b=NULL) {
      if (from_b==NULL) {
           b=bam_init1();
           novel=true;
           }
        else {
           b=from_b;
           novel=false;
           }
      }
   //creates a new record from 1-based alignment coordinate
   //quals should be given as Phred33
   GBamRecord(const char* qname, int32_t gseq_tid,
                    int pos, bool reverse, const char* qseq, const char* cigar=NULL, const char* quals=NULL) {
      novel=true;
      b=bam_init1();
      b->core.tid=gseq_tid;
      if (pos<=0) b->core.pos=-1; //unmapped?
             else b->core.pos=pos-1; //BAM is 0-based
      b->core.qual=255;
      int l_qseq=strlen(qseq);
      //this may not be accurate, setting CIGAR is the correct way
      //b->core.bin = bam_reg2bin(b->core.pos, b->core.pos+l_qseq-1);
      b->core.l_qname=strlen(qname)+1; //includes the \0 at the end
      memcpy(realloc_bdata(b, b->core.l_qname), qname, b->core.l_qname);
      set_cigar(cigar); //this will also set core.bin
      add_sequence(qseq, l_qseq);
      add_quals(quals); //quals must be given as Phred33
      if (reverse) { b->core.flag |= BAM_FREVERSE ; }
      }

    void set_mdata(int32_t mtid, int32_t mpos, 
                     int32_t isize=0) { //mate info for current record
      b->core.mtid=mtid;
      b->core.mpos=mpos; // should be -1 if '*'
      b->core.isize=isize; //should be 0 if not available
      }

    void set_flags(uint16_t flags) {
      b->core.flag=flags;
      }

    void set_cigar(const char* cigar) {
      //requires b->core.pos to have been set properly PRIOR to this call
      int doff=b->core.l_qname;
      uint8_t* after_cigar=NULL;
      int after_cigar_len=0;
      uint8_t* prev_bdata=NULL;
      if (b->data_len>doff) {
         //cigar string already allocated, replace it
         int d=b->core.l_qname + b->core.n_cigar * 4;//offset of after-cigar data
         after_cigar=b->data+d;
         after_cigar_len=b->data_len-d;
         }
      const char *s;
      char *t;
      int i, op;
      long x;
      b->core.n_cigar = 0;
      if (cigar!=NULL && cigar[0] != '*') {
           for (s = cigar; *s; ++s) {
               if (isalpha(*s)) b->core.n_cigar++;
               else if (!isdigit(*s)) {
                    fprintf(stderr, "Error: invalid CIGAR character (%s)\n",cigar);
                    exit(1);
                    }
               }
           if (after_cigar_len>0) { //replace/insert into existing full data
                prev_bdata=dupalloc_bdata(b, doff + b->core.n_cigar * 4 + after_cigar_len);
                memcpy((void*)(b->data+doff+b->core.n_cigar*4),(void*)after_cigar, after_cigar_len);
                free(prev_bdata);
                }
              else {
                b->data = realloc_bdata(b, doff + b->core.n_cigar * 4);
                }
           for (i = 0, s = cigar; i != b->core.n_cigar; ++i) {
               x = strtol(s, &t, 10);
               op = toupper(*t);
               if (op == 'M' || op == '=' || op == 'X') op = BAM_CMATCH;
               else if (op == 'I') op = BAM_CINS;
               else if (op == 'D') op = BAM_CDEL;
               else if (op == 'N') op = BAM_CREF_SKIP;
               else if (op == 'S') op = BAM_CSOFT_CLIP;
               else if (op == 'H') op = BAM_CHARD_CLIP;
               else if (op == 'P') op = BAM_CPAD;
               else { fprintf(stderr, "Error: invalid CIGAR operation (%s)\n",cigar);
                      exit(1);
                     }
               s = t + 1;
               bam1_cigar(b)[i] = x << BAM_CIGAR_SHIFT | op;
           }
           if (*s) {fprintf(stderr, "Error: unmatched CIGAR operation (%s)\n",cigar); exit(1); }
           b->core.bin = bam_reg2bin(b->core.pos, bam_calend(&b->core, bam1_cigar(b)));
       } else {//no CIGAR string
           if (!(b->core.flag&BAM_FUNMAP)) {
               fprintf(stderr, "Warning: mapped sequence without CIGAR (%s)\n", (char*)b->data);
               b->core.flag |= BAM_FUNMAP;
           }
           b->core.bin = bam_reg2bin(b->core.pos, b->core.pos + 1);
       }
      } //set_cigar()

    void add_sequence(const char* qseq, int slen=-1) {
      //must be called AFTER set_cigar (cannot replace existing sequence for now)
      if (qseq==NULL) return; //should we ever care about this?
      if (slen<0) slen=strlen(qseq);
      int doff = b->core.l_qname + b->core.n_cigar * 4;
      if (strcmp(qseq, "*")!=0) {
          b->core.l_qseq=slen;
          if (b->core.n_cigar && b->core.l_qseq != (int32_t)bam_cigar2qlen(&b->core, bam1_cigar(b)))
              {
              fprintf(stderr, "Error: CIGAR and sequence length are inconsistent!(%s)\n", qseq);
              exit(1);
              }
          uint8_t* p = (uint8_t*)realloc_bdata(b, doff + (b->core.l_qseq+1)/2 + b->core.l_qseq) + doff;
          //also allocated quals memory
          memset(p, 0, (b->core.l_qseq+1)/2);
          for (int i = 0; i < b->core.l_qseq; ++i)
              p[i/2] |= bam_nt16_table[(int)qseq[i]] << 4*(1-i%2);
          } else b->core.l_qseq = 0;
      }

    void add_quals(const char* quals) {
      //requires core.l_qseq already set
      //and must be called AFTER add_sequence(), which also allocates the memory for quals
      uint8_t* p = b->data+(b->core.l_qname + b->core.n_cigar * 4 + (b->core.l_qseq+1)/2);
      if (quals==NULL || quals[0]=='*') {
         for (int i=0;i < b->core.l_qseq; i++) p[i] = 0xff;
         return;
         }
      for (int i=0;i < b->core.l_qseq; i++) p[i] = quals[i]-33;
      }
    void clear() {
       if (novel) {
           bam_destroy1(b);
           }
       novel=true;
       b=bam_init1();
       }
   ~GBamRecord() {
      if (novel) {  bam_destroy1(b); }
      }
   void parse_error(const char* s) {
     fprintf(stderr, "Parsing error: %s\n", s);
     exit(1);
     }

   void add_aux(const char tag[2], char atype, int len, uint8_t *data) {
      int ori_len = b->data_len;
      b->data_len += 3 + len;
      b->l_aux += 3 + len;
      if (b->m_data < b->data_len) {
        b->m_data = b->data_len;
        kroundup32(b->m_data);
        b->data = (uint8_t*)realloc(b->data, b->m_data);
      }
      b->data[ori_len] = tag[0]; b->data[ori_len + 1] = tag[1];
      b->data[ori_len + 2] = atype;
      memcpy(b->data + ori_len + 3, data, len);
      }

   void add_aux(char* str) {
      static char tag[2];
      static uint8_t abuf[512];
      //requires: being called AFTER add_quals()
      int strl=strlen(str);
      //int doff = b->core.l_qname + b->core.n_cigar*4 + (b->core.l_qseq+1)/2 + b->core.l_qseq + b->l_aux;
      //int doff0=doff;
      if (strl < 6 || str[2] != ':' || str[4] != ':')
          parse_error("missing colon in auxiliary data");
      tag[0] = str[0]; tag[1] = str[1];
      uint8_t atype = str[3];
      uint8_t* adata=abuf;
      int alen=0;
      if (atype == 'A' || atype == 'a' || atype == 'c' || atype == 'C') { // c and C for backward compatibility
          atype='A';
          alen=1;
          adata=(uint8_t*)&str[5];
          }
       else if (atype == 'I' || atype == 'i') {
          long long x=(long long)atoll(str + 5);
          if (x < 0) {
              if (x >= -127) {
                  atype='c';
                  abuf[0] =  (int8_t)x;
                  alen=1;
                  }
              else if (x >= -32767) {
                  atype = 's';
                  *(int16_t*)abuf = (int16_t)x;
                  alen=2;
                  }
              else {
                  atype='i';
                  *(int32_t*)abuf = (int32_t)x;
                  alen=4;
                  if (x < -2147483648ll)
                      fprintf(stderr, "Parse warning: integer %lld is out of range.",
                              x);
                  }
              } else { //x >=0
              if (x <= 255) {
                  atype = 'C';
                  abuf[0] = (uint8_t)x;
                  alen=1;
                  }
              else if (x <= 65535) {
                  atype='S';
                  *(uint16_t*)abuf = (uint16_t)x;
                  alen=2;
                  }
              else {
                  atype='I';
                  *(uint32_t*)abuf = (uint32_t)x;
                  alen=4;
                  if (x > 4294967295ll)
                      fprintf(stderr, "Parse warning: integer %lld is out of range.",
                              x);
                  }
              }
          } //integer type
          else if (atype == 'f') {
              *(float*)abuf = (float)atof(str + 5);
              alen = sizeof(float);
              }
          else if (atype == 'd') { //?
              *(float*)abuf = (float)atof(str + 9);
              alen=8;
              }
          else if (atype == 'Z' || atype == 'H') {
              if (atype == 'H') { // check whether the hex string is valid
                  if ((strl - 5) % 2 == 1) parse_error("length of the hex string not even");
                  for (int i = 0; i < strl - 5; ++i) {
                      int c = toupper(str[5 + i]);
                      if (!((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F')))
                          parse_error("invalid hex character");
                      }
                  }
              memcpy(abuf, str + 5, strl - 5);
              abuf[strl-5] = 0;
              alen=strl-4;
              } else parse_error("unrecognized aux type");
   add_aux(tag, atype, alen, adata);
   }//add_aux()

   bam1_t* get_b() { return b; }
   

};

class GBamWriter {
   samfile_t* bam_file;
   bam_header_t* bam_header;
 public:
   void create(const char* fname, bool uncompressed=false) {
      if (bam_header==NULL) {
         fprintf(stderr, "Error: no bam_header for GBamWriter::create()!\n");
         exit(1);
         }
      if (uncompressed) {
         bam_file=samopen(fname, "wbu", bam_header);
         }
        else {
         bam_file=samopen(fname, "wb", bam_header);
         }
      if (bam_file==NULL) {
         fprintf(stderr, "Error: could not create BAM file %s!\n",fname);
         exit(1);
         }
      //do we need to call bam_header_write() ?
      }

   GBamWriter(const char* fname, bam_header_t* bh, bool uncompressed=false) {
      bam_header=bh;
      create(fname, uncompressed);
      }

   GBamWriter(const char* fname, const char* samfname, bool uncompressed=false) {
      tamFile samf_in=sam_open(samfname);
      if (samf_in==NULL) {
         fprintf(stderr, "Error: could not open SAM file %s\n", samfname);
         exit(1);
         }
      bam_header=sam_header_read(samf_in);
      if (bam_header==NULL) {
         fprintf(stderr, "Error: could not read SAM header from %s!\n",samfname);
         exit(2);
         }
      sam_close(samf_in);
      create(fname, uncompressed);
      }
      
    ~GBamWriter() {
      samclose(bam_file);
      bam_header_destroy(bam_header);
      }
   bam_header_t* get_header() { return bam_header; }
   int32_t get_tid(const char *seq_name) {
      if (bam_header==NULL) {
         fprintf(stderr, "Error: missing SAM header (get_tid())\n");
         exit(1);
         }
      return bam_get_tid(bam_header, seq_name);
      }

   //just a convenience function for creating a new record, but it's NOT written
   //given pos must be 1-based (so it'll be stored as pos-1 because BAM is 0-based)
   GBamRecord* new_record(const char* qname, const char* gseqname,
            int pos, bool reverse, const char* qseq, const char* cigar=NULL, const char* qual=NULL) {
      int32_t gseq_tid=get_tid(gseqname);
      if (gseq_tid < 0 && strcmp(gseqname, "*")) {
            if (bam_header->n_targets == 0) {
               fprintf(stderr, "Error: missing/invalid SAM header\n");
               exit(1);
               } else
                   fprintf(stderr, "Warning: reference '%s' not found in header, will consider it '*'.\n",
                                   gseqname);
            }

      GBamRecord *newbamrec=new GBamRecord(qname, gseq_tid, pos, reverse, qseq, cigar, qual);
      return newbamrec;
      }

   void write(GBamRecord* brec) {
      samwrite(this->bam_file,brec->get_b());
      }
};

int str_split(char* str, char**& t, char delim='\t') {
  //split text line str into an array of strings t[] by delimiter delim (tab)
  //Notes: * destructive operation for s (replaces every \t with \0)
  //       * user must free t when no longer needed
  int tcap=14;
  t=(char**)malloc(tcap*sizeof(char*));
  int tcount=0;
  char prevch=0;
  for (char* p = str; *p!=0 ;p++) {
    if (*p==delim) *p=0; //break the string here
    if (prevch==0) { //field start
      if (tcount==tcap) {
         tcap+=4;
         t = (char**)realloc(t,tcap*sizeof(char*));
         }
      t[tcount]=p;
      tcount++;
      } //field start
    prevch=*p;
    if (*p=='\n' || *p=='\r') {
        *p=0;
        break;
        }
    }//for each character on the line
  return tcount;
}

int main(int argc, char *argv[]) {
 
  if (argv[1]==NULL) {
     fprintf(stderr, "Usage:\nsam2bam <samfile>\n");
     exit(1);
     }
  string fname(argv[1]);
  fname+=".bam";
  GBamWriter wbam(fname.c_str(), argv[1]);
  //now parse the SAM line and convert it to BAM
  char samline[1024];
  samline[1023]=0;
  FILE* f_in=fopen(argv[1], "r");
  if (f_in==NULL) {
     fprintf(stderr, "Error: could not open SAM file %s for reading!\n",argv[1]);
     exit(2);
     }
  while (fgets(samline, 1024, f_in)!=NULL) {
    char** t=NULL;
    int tcount=str_split(samline, t);
    if (samline[0]=='@') {
       free(t);
       continue; //skip header
       }
    if (tcount<10) {
       fprintf(stderr, "Warning: skipping malformed SAM line %s\n",samline);
       free(t);
       continue;
       }
    int gpos=isdigit(t[3][0]) ? atoi(t[3]) : 0;
    GBamRecord *brec= wbam.new_record(t[0],t[2],gpos,false, t[9], t[5], t[10]);
    //parse flags value:
    char* s=NULL;
    long flag = strtol(t[1], &s, 0);
    if (*s) { // not the end of the string
        flag = 0;
        for (s = t[1]; *s; ++s)
            flag |= bam_char2flag_table[(int)*s]; //text-style SAM flags
        }
    brec->set_flags(flag);
    int32_t mtid = strcmp(t[6], "=")? wbam.get_tid(t[6]) : brec->get_b()->core.tid;
    int32_t mpos = isdigit(t[7][0])? atoi(t[7]) - 1 : -1;
    int32_t isize = (t[8][0] == '-' || isdigit(t[8][0]))? atoi(t[8]) : 0;
    brec->set_mdata(mtid, mpos, isize);
    //now parse and add aux data:
    if (tcount>11) {
      for (int i=11;i<tcount;i++) {
         //parse aux entry:
         brec->add_aux(t[i]);
         }//for each aux field
      }
    wbam.write(brec);
    delete brec;
    free(t);
   }
  fclose(f_in);
  return 0;
}
